/*
Copyright 2017-2018 Aron Barath

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#include "ncbricks.h"
#include <stdlib.h>
#include <string.h>

static void ball_to_disp(struct ball* const ball)
{
	ball->dpos.x = ball->lpos.x / LOGICAL_SCALE;
	ball->dpos.y = ball->lpos.y / LOGICAL_SCALE;
}

static void delete_brick(struct brick* const brick, const size_t index)
{
	if(brick->charm)
	{
		charm_spawn(brick, brick->charm);
	}

	ptrarr_remove(&g_bricks, index);
	brick_hide(brick);
	free(brick);
}

static bool animate_ball(struct ball* const ball, const struct rect2* const paddle, const int max_x, const int max_y)
{
	vec2_add(&ball->lpos, &ball->lpos, &ball->velo);

	if(0>ball->lpos.x)
	{
		ball->lpos.x = -ball->lpos.x;
		ball->velo.x = -ball->velo.x;
	}

	if(max_x<=ball->lpos.x)
	{
		ball->lpos.x = max_x*2-ball->lpos.x;
		ball->velo.x = -ball->velo.x;
	}

	if(0>ball->lpos.y)
	{
		ball->lpos.y = -ball->lpos.y;
		ball->velo.y = -ball->velo.y;
	}

	if(rect2_contains(paddle, &ball->lpos))
	{
		ball_reflect(ball, paddle);
	}

	ball_to_disp(ball);

	return (max_y<=ball->lpos.y);
}

static void brick_to_log(struct rect2* const lrect, const struct rect2* const drect)
{
	lrect->pos.x = drect->pos.x * LOGICAL_SCALE;
	lrect->pos.y = drect->pos.y * LOGICAL_SCALE;
	lrect->end.x = drect->end.x * LOGICAL_SCALE + LOGICAL_SCALE - 1;
	lrect->end.y = drect->end.y * LOGICAL_SCALE + LOGICAL_SCALE - 1;
}

static void animate_ball_bricks(struct ball* const ball, const struct vec2* opos)
{
	size_t i;
	const struct vec2* const npos = &ball->lpos;

	struct vec2         new_opos;
	size_t              cbrickV; /* index of the contacting brick */
	size_t              cbrickH;
	struct vec2         isect; /* temporary intersection point */
	struct vec2         ctvV; /* contact vectors */
	struct vec2         ctvH;
	struct vec2         ctpV; /* contact point */
	struct vec2         ctpH;
	int                 ctdepthV; /* contact depths (L1 norm, smaller is better) */
	int                 ctdepthH;
	bool                contactV; /* there is a contact */
	bool                contactH;

again:

	contactH = FALSE;
	contactV = FALSE;

	for(i=0;i<g_bricks.size;++i)
	{
		struct brick* const brick = (struct brick*)g_bricks.pointers[i];
		struct rect2        rec; /* the brick rect in logical space */

		brick_to_log(&rec, &brick->pos);

		/* top-side (vertical rebounce) */

		if(/*opos->y>rec.pos.y && rec.pos.y>=npos->y &&*/
			0==vec2_intersect_vr(&isect, opos, npos,
			rec.pos.x, rec.pos.y, rec.end.x, rec.pos.y))
		{
			const int len = vec2_L1norm(opos, &isect);

			if(!contactV || len<ctdepthV)
			{
				ctdepthV = len;
				memcpy(&ctpV, &isect, sizeof(isect));
				vec2_sub(&ctvV, &isect, opos);

				cbrickV  = i;
				contactV = TRUE;
			}
		}

		/* bottom-side (vertical rebounce) */

		if(/*opos->y<rec.end.y && rec.end.y<=npos->y &&*/
			0==vec2_intersect_vr(&isect, opos, npos,
			rec.pos.x, rec.end.y, rec.end.x, rec.end.y))
		{
			const int len = vec2_L1norm(opos, &isect);

			if(!contactV || len<ctdepthV)
			{
				ctdepthV = len;
				memcpy(&ctpV, &isect, sizeof(isect));
				vec2_sub(&ctvV, &isect, opos);

				cbrickV  = i;
				contactV = TRUE;
			}
		}

		/* left-side (horizontal rebounce) */

		if(/*opos->x<rec.pos.x && rec.pos.x<=npos->x &&*/
			0==vec2_intersect_vr(&isect, opos, npos,
			rec.pos.x, rec.pos.y, rec.pos.x, rec.end.y))
		{
			const int len = vec2_L1norm(opos, &isect);

			if(!contactH || len<ctdepthH)
			{
				ctdepthH = len;
				memcpy(&ctpH, &isect, sizeof(isect));
				vec2_sub(&ctvH, &isect, opos);

				cbrickH  = i;
				contactH = TRUE;
			}
		}

		/* right-side (horizontal rebounce) */

		if(/*opos->x>rec.end.x && rec.end.x>=npos->x &&*/
			0==vec2_intersect_vr(&isect, opos, npos,
			rec.end.x, rec.pos.y, rec.end.x, rec.end.y))
		{
			const int len = vec2_L1norm(opos, &isect);

			if(!contactH || len<ctdepthH)
			{
				ctdepthH = len;
				memcpy(&ctpH, &isect, sizeof(isect));
				vec2_sub(&ctvH, &isect, opos);

				cbrickH  = i;
				contactH = TRUE;
			}
		}
	}

	if(contactH || contactV)
	{
		struct brick* brick;
		size_t        cbrick;
		bool          delete = FALSE;
		bool          bounce = FALSE;

		if(contactV && (!contactH || (contactH && ctdepthV<ctdepthH)))
		{
			cbrick   = cbrickV;
			contactH = FALSE;
		}
		else
		{
			cbrick   = cbrickH;
			contactV = FALSE;
		}

		brick = (struct brick*)g_bricks.pointers[cbrick];

		/* decrease hit points in need */

		if(0<brick->hitpts)
		{
			/* reward the player (even more if more hits since last paddle touch) */

			add_score(SCORE_BRICK * (++ball->hits));

			if(0==(--brick->hitpts))
			{
				/* brick is broken */
				delete = TRUE;
				bounce = !g_atomicball;
			}
			else
			{
				/* brick still have hit points */
				brick_show(brick);
				bounce = TRUE;
			}
		}
		else
		{
			/* unbreakable brick, just bounce */
			bounce = TRUE;
		}

		if(bounce)
		{
			if(contactV)
			{
				ctvV.y = -ctvV.y;
				vec2_add(&ball->lpos, &ctpV, &ctvV);
				ball->velo.y = -ball->velo.y;

				memcpy(&new_opos, &ctpV, sizeof(ctpV));

				if(0>ball->velo.y)
				{
					--new_opos.y;
					--ball->lpos.y;
				}
				else
				{
					++new_opos.y;
					++ball->lpos.y;
				}
			}
			else
			{
				ctvH.x = -ctvH.x;
				vec2_add(&ball->lpos, &ctpH, &ctvH);
				ball->velo.x = -ball->velo.x;

				memcpy(&new_opos, &ctpH, sizeof(ctpH));

				if(0>ball->velo.x)
				{
					--new_opos.x;
					--ball->lpos.x;
				}
				else
				{
					++new_opos.x;
					++ball->lpos.x;
				}
			}

			opos = &new_opos;
			ball_to_disp(ball);
		}

		if(delete)
		{
			delete_brick(brick, cbrick);
		}

		goto again;
	}
}

static bool animate_missile(struct missile* const missile)
{
	size_t i;

	if(0>(--missile->pos.y))
	{
		return TRUE;
	}

	for(i=0;i<g_bricks.size;++i)
	{
		struct brick* const brick = (struct brick*)g_bricks.pointers[i];

		if(rect2_contains(&brick->pos, &missile->pos))
		{
			if(0<brick->hitpts)
			{
				/* reward the player (same amount of point each time!) */

				add_score(SCORE_BRICK);

				/* decrease brick hit points by one, delete if needed */

				if(0==(--brick->hitpts))
				{
					delete_brick(brick, i--);
				}
				else
				{
					brick_show(brick);
				}
			}

			return TRUE;
		}
	}

	return FALSE;
}

static bool animate_charm(struct charm* const charm, const struct paddle* const paddle)
{
	size_t i;

	for(i=0;i<g_bricks.size;++i)
	{
		struct brick* const brick = (struct brick*)g_bricks.pointers[i];

		if(rect2_contains(&brick->pos, &charm->pos))
		{
			brick_show(brick);
			break;
		}
	}

	if(g_paddle_y==(++charm->pos.y) && paddle->x<=charm->pos.x && charm->pos.x<=(paddle->x+paddle->width-1))
	{
		charm->effect();
		return TRUE;
	}

	return (g_paddle_y<=charm->pos.y);
}

void animate_all(bool anim_balls)
{
	const int max_x = LOGICAL_SCALE * g_board_width - 1;
	const int max_y = LOGICAL_SCALE * g_board_height - 1;

	size_t i;

	missile_hide_all();
	charm_hide_all();

	if(anim_balls)
	{
		ball_hide_all();
		for(i=0;i<g_balls.size;++i)
		{
			struct ball* const ball = (struct ball*)g_balls.pointers[i];
			struct vec2 old_pos;

			memcpy(&old_pos, &ball->lpos, sizeof(old_pos));

			if(animate_ball(ball, &g_paddle.lrect, max_x, max_y))
			{
				ptrarr_remove(&g_balls, i--);
				free(ball);
				continue;
			}

			animate_ball_bricks(ball, &old_pos);
		}
		ball_show_all();
	}

	for(i=0;i<g_missiles.size;++i)
	{
		struct missile* const missile = (struct missile*)g_missiles.pointers[i];

		if(animate_missile(missile))
		{
			ptrarr_remove(&g_missiles, i--);
			free(missile);
			continue;
		}
	}

	for(i=0;i<g_charms.size;++i)
	{
		struct charm* const charm = (struct charm*)g_charms.pointers[i];

		if(animate_charm(charm, &g_paddle))
		{
			ptrarr_remove(&g_charms, i--);
			free(charm);
			continue;
		}
	}

	missile_show_all();
	charm_show_all();
}
